@@ -16,6 +16,7 @@
 # Requires Python 2.4+ and Openssl 1.0+
 #
 
+import multiprocessing
 import os
 import re
 import shutil
@@ -50,7 +51,8 @@ class DefaultOSUtil(object):
 
     def __init__(self):
         self.agent_conf_file_path = '/etc/waagent.conf'
-        self.selinux=None
+        self.selinux = None
+        self.disable_route_warning = False
 
     def get_agent_conf_file_path(self):
         return self.agent_conf_file_path
@@ -247,16 +249,6 @@ class DefaultOSUtil(object):
         else:
             return False
 
-    def set_selinux_enforce(self, state):
-        """
-        Calls shell command 'setenforce' with 'state'
-        and returns resulting exit code.
-        """
-        if self.is_selinux_system():
-            if state: s = '1'
-            else: s='0'
-            return shellutil.run("setenforce "+s)
-
     def set_selinux_context(self, path, con):
         """
         Calls shell 'chcon' with 'path' and 'con' context.
@@ -447,7 +439,8 @@ class DefaultOSUtil(object):
             iface=sock[i:i+16].split(b'\0', 1)[0]
             if len(iface) == 0 or self.is_loopback(iface) or iface != primary:
                 # test the next one
-                logger.info('interface [{0}] skipped'.format(iface))
+                if len(iface) != 0 and not self.disable_route_warning:
+                    logger.info('interface [{0}] skipped'.format(iface))
                 continue
             else:
                 # use this one
@@ -479,7 +472,8 @@ class DefaultOSUtil(object):
         primary = None
         primary_metric = None
 
-        logger.info("examine /proc/net/route for primary interface")
+        if not self.disable_route_warning:
+            logger.info("examine /proc/net/route for primary interface")
         with open('/proc/net/route') as routing_table:
             idx = 0
             for header in filter(lambda h: len(h) > 0, routing_table.readline().strip(" \n").split("\t")):
@@ -503,11 +497,18 @@ class DefaultOSUtil(object):
 
         if primary is None:
             primary = ''
-
-        logger.info('primary interface is [{0}]'.format(primary))
+            if not self.disable_route_warning:
+                with open('/proc/net/route') as routing_table_fh:
+                    routing_table_text = routing_table_fh.read()
+                    logger.error('could not determine primary interface, '
+                                 'please ensure /proc/net/route is correct:\n'
+                                 '{0}'.format(routing_table_text))
+                    self.disable_route_warning = True
+        else:
+            logger.info('primary interface is [{0}]'.format(primary))
+            self.disable_route_warning = False
         return primary
 
-
     def is_primary_interface(self, ifname):
         """
         Indicate whether the specified interface is the primary.
@@ -516,13 +517,14 @@ class DefaultOSUtil(object):
         """
         return self.get_primary_interface() == ifname
 
-
     def is_loopback(self, ifname):
         s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
         result = fcntl.ioctl(s.fileno(), 0x8913, struct.pack('256s', ifname[:15]))
         flags, = struct.unpack('H', result[16:18])
         isloopback = flags & 8 == 8
-        logger.info('interface [{0}] has flags [{1}], is loopback [{2}]'.format(ifname, flags, isloopback))
+        if not self.disable_route_warning:
+            logger.info('interface [{0}] has flags [{1}], '
+                        'is loopback [{2}]'.format(ifname, flags, isloopback))
         return isloopback
 
     def get_dhcp_lease_endpoint(self):
@@ -649,7 +651,7 @@ class DefaultOSUtil(object):
         return shellutil.run(cmd, chk_err=False)
 
     def get_dhcp_pid(self):
-        ret= shellutil.run_get_output("pidof dhclient")
+        ret = shellutil.run_get_output("pidof dhclient", chk_err=False)
         return ret[1] if ret[0] == 0 else None
 
     def set_hostname(self, hostname):
@@ -684,6 +686,7 @@ class DefaultOSUtil(object):
 
     def publish_hostname(self, hostname):
         self.set_dhcp_hostname(hostname)
+        self.set_hostname_record(hostname)
         ifname = self.get_if_name()
         self.restart_if(ifname)
 
@@ -734,22 +737,39 @@ class DefaultOSUtil(object):
             port_id = port_id - 2
         device = None
         path = "/sys/bus/vmbus/devices/"
-        for vmbus in os.listdir(path):
-            deviceid = fileutil.read_file(os.path.join(path, vmbus, "device_id"))
-            guid = deviceid.lstrip('{').split('-')
-            if guid[0] == g0 and guid[1] == "000" + ustr(port_id):
-                for root, dirs, files in os.walk(path + vmbus):
-                    if root.endswith("/block"):
-                        device = dirs[0]
-                        break
-                    else : #older distros
-                        for d in dirs:
-                            if ':' in d and "block" == d.split(':')[0]:
-                                device = d.split(':')[1]
-                                break
-                break
+        if os.path.exists(path):
+            for vmbus in os.listdir(path):
+                deviceid = fileutil.read_file(os.path.join(path, vmbus, "device_id"))
+                guid = deviceid.lstrip('{').split('-')
+                if guid[0] == g0 and guid[1] == "000" + ustr(port_id):
+                    for root, dirs, files in os.walk(path + vmbus):
+                        if root.endswith("/block"):
+                            device = dirs[0]
+                            break
+                        else : #older distros
+                            for d in dirs:
+                                if ':' in d and "block" == d.split(':')[0]:
+                                    device = d.split(':')[1]
+                                    break
+                    break
         return device
 
+    def set_hostname_record(self, hostname):
+        fileutil.write_file(conf.get_published_hostname(), contents=hostname)
+
+    def get_hostname_record(self):
+        hostname_record = conf.get_published_hostname()
+        if not os.path.exists(hostname_record):
+            # this file is created at provisioning time with agents >= 2.2.3
+            hostname = socket.gethostname()
+            logger.warn('Hostname record does not exist, '
+                        'creating [{0}] with hostname [{1}]',
+                        hostname_record,
+                        hostname)
+            self.set_hostname_record(hostname)
+        record = fileutil.read_file(hostname_record)
+        return record
+
     def del_account(self, username):
         if self.is_sys_user(username):
             logger.error("{0} is a system user. Will not delete it.", username)
@@ -758,22 +778,14 @@ class DefaultOSUtil(object):
         self.conf_sudoer(username, remove=True)
 
     def decode_customdata(self, data):
-        return base64.b64decode(data)
+        return base64.b64decode(data).decode('utf-8')
 
     def get_total_mem(self):
-        cmd = "grep MemTotal /proc/meminfo |awk '{print $2}'"
-        ret = shellutil.run_get_output(cmd)
-        if ret[0] == 0:
-            return int(ret[1])/1024
-        else:
-            raise OSUtilError("Failed to get total memory: {0}".format(ret[1]))
+        # Get total memory in bytes and divide by 1024**2 to get the value in MB.
+        return os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES') / (1024**2)
 
     def get_processor_cores(self):
-        ret = shellutil.run_get_output("grep 'processor.*:' /proc/cpuinfo |wc -l")
-        if ret[0] == 0:
-            return int(ret[1])
-        else:
-            raise OSUtilError("Failed to get processor cores")
+        return multiprocessing.cpu_count()
 
     def set_admin_access_to_ip(self, dest_ip):
         #This allows root to access dest_ip
